class MouseEvent(object):
    def __init__(self, event, x, y, flags, param):
        self.event = event
        self.x = x
        self.y = y
        self.flags = flags
        self.param = param

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "event:{}\nx,y:{},{}\nflags:{}\nparam:{}\n".format(self.event, self.x, self.y, self.flags, self.param)



import numpy as np
from collections import Hashable


def uid_for_source(video_source):
    if len(str(video_source)) <= 1000:
        uid = str(video_source)
    elif isinstance(video_source, Hashable):
        uid = str(hash(video_source))
    else:
        uid = str(hash(str(video_source)))
    return uid



import warnings

import cv2
import numpy as np

from .winctrl import WinCtrl
from cvpubsubs.webcam_pub.camctrl import CamCtrl
from cvpubsubs.webcam_pub.frame_handler import VideoHandlerThread
from localpubsub import NoData
from cvpubsubs.window_sub.mouse_event import MouseEvent
from cvpubsubs.serialize import uid_for_source

from typing import List, Union, Callable, Any, Dict
import numpy as np
from cvpubsubs.callbacks import global_cv_display_callback


class SubscriberWindows(object):
    frame_dict = {}

    esc_key_codes = [27]  # ESC key on most keyboards

    def __init__(self,
                 window_names=('cvpubsubs',),  # type: List[str]
                 video_sources=(0,),  # type: List[Union[str,int]]
                 callbacks=(None,),  # type: List[Callable[[List[np.ndarray]], Any]]
                 ):
        self.source_names = []
        self.close_threads = None
        self.frames = []
        self.input_vid_global_names = []
        self.window_names = []
        self.input_cams = []

        for name in video_sources:
            self.add_source(name)
        self.callbacks = callbacks
        for name in window_names:
            self.add_window(name)

    def add_source(self, name):
        uid = uid_for_source(name)
        self.source_names.append(uid)
        self.input_vid_global_names.append(uid + "frame")
        self.input_cams.append(name)

    def add_window(self, name):
        self.window_names.append(name)
        cv2.namedWindow(name + " (press ESC to quit)")
        cv2.setMouseCallback(name + " (press ESC to quit)", self.handle_mouse)

    def add_callback(self, callback):
        self.callbacks.append(callback)

    @staticmethod
    def set_global_frame_dict(name, *args):
        if len(str(name)) <= 1000:
            SubscriberWindows.frame_dict[str(name) + "frame"] = list(args)
        elif isinstance(name, np.ndarray):
            SubscriberWindows.frame_dict[str(hash(str(name))) + "frame"] = list(args)
        else:
            raise ValueError("Input window name too long.")

    def __stop_all_cams(self):
        for c in self.source_names:
            CamCtrl.stop_cam(c)

    def handle_keys(self,
                    key_input,  # type: int
                    ):
        if key_input in self.esc_key_codes:
            for name in self.window_names:
                cv2.destroyWindow(name + " (press ESC to quit)")
            WinCtrl.quit()
            self.__stop_all_cams()
            return 'quit'
        elif key_input not in [-1, 0]:
            try:
                WinCtrl.key_pub.publish(chr(key_input))
            except ValueError:
                warnings.warn(
                    RuntimeWarning("Unknown key code: [{}]. Please report to cv_pubsubs issue page.".format(key_input))
                )

    def handle_mouse(self, event, x, y, flags, param):
        mousey = MouseEvent(event, x, y, flags, param)
        WinCtrl.mouse_pub.publish(mousey)

    def _display_frames(self, frames, win_num, ids=None):
        if isinstance(frames, Exception):
            raise frames
        for f in range(len(frames)):
            # detect nested:
            if isinstance(frames[f], (list, tuple)) or frames[f].dtype.num == 17 or len(frames[f].shape) > 3:
                win_num = self._display_frames(frames[f], win_num, ids)
            else:
                cv2.imshow(self.window_names[win_num % len(self.window_names)] + " (press ESC to quit)", frames[f])
                win_num += 1
        return win_num

    def update_window_frames(self):
        win_num = 0
        for i in range(len(self.input_vid_global_names)):
            if self.input_vid_global_names[i] in self.frame_dict and \
                    not isinstance(self.frame_dict[self.input_vid_global_names[i]], NoData):
                if len(self.callbacks) > 0 and self.callbacks[i % len(self.callbacks)] is not None:
                    self.frames = self.callbacks[i % len(self.callbacks)](
                        self.frame_dict[self.input_vid_global_names[i]])
                else:
                    self.frames = self.frame_dict[self.input_vid_global_names[i]]
                if isinstance(self.frames, np.ndarray) and len(self.frames.shape) <= 3:
                    self.frames = [self.frames]
                win_num = self._display_frames(self.frames, win_num)

    def update(self, arr=None, id=None):
        if arr is not None and id is not None:
            global_cv_display_callback(arr, id)
            if id not in self.input_cams:
                self.add_source(id)
                self.add_window(id)
        sub_cmd = WinCtrl.win_cmd_sub()
        self.update_window_frames()
        msg_cmd = sub_cmd.get()
        key = self.handle_keys(cv2.waitKey(1))
        return msg_cmd, key

    def wait_for_init(self):
        msg_cmd=""
        key = ""
        while msg_cmd != 'quit' and key != 'quit' and len(self.frames)==0:
            msg_cmd, key = self.update()

    def end(self):
        if self.close_threads is not None:
            for t in self.close_threads:
                t.join()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end()

    # todo: figure out how to get the red x button to work. Try: https://stackoverflow.com/a/37881722/782170
    def loop(self):
        sub_cmd = WinCtrl.win_cmd_sub()
        msg_cmd = ''
        key = ''
        while msg_cmd != 'quit' and key != 'quit':
            msg_cmd, key = self.update()
        sub_cmd.release()
        WinCtrl.quit(force_all_read=False)
        self.__stop_all_cams()


from cvpubsubs.callbacks import global_cv_display_callback

from threading import Thread


def display(*vids,
            callbacks: Union[Dict[Any, Callable], List[Callable], Callable, None] = None,
            window_names=[],
            blocking=False):
    vid_threads = []
    if isinstance(callbacks, Dict):
        for v in vids:
            v_name = uid_for_source(v)
            v_callbacks = []
            if v_name in callbacks:
                v_callbacks.extend(callbacks[v_name])
            if v in callbacks:
                v_callbacks.extend(callbacks[v])
            vid_threads.append(VideoHandlerThread(v, callbacks=v_callbacks))
    elif isinstance(callbacks, List):
        for v in vids:
            vid_threads.append(VideoHandlerThread(v, callbacks=callbacks))
    elif isinstance(callbacks, Callable):
        for v in vids:
            vid_threads.append(VideoHandlerThread(v, callbacks=[callbacks]))
    else:
        for v in vids:
            vid_threads.append(VideoHandlerThread(v))
    for v in vid_threads:
        v.start()
    if len(window_names) == 0:
        window_names = ["window {}".format(i) for i in range(len(vids))]
    if blocking:
        SubscriberWindows(window_names=window_names,
                          video_sources=vids
                          ).loop()
        for v in vid_threads:
            v.join()
    else:
        s = SubscriberWindows(window_names=window_names,
                              video_sources=vids
                              )
        s.close_threads = vid_threads
        v_names = []
        for v in vids:
            v_name = uid_for_source(v)
            v_names.append(v_name)
        return s, v_names



from cvpubsubs.window_sub.winctrl import WinCtrl
import threading
import time

if False:
    from typing import Callable
    from cvpubsubs.window_sub.mouse_event import MouseEvent


class mouse_thread(object):  # NOSONAR

    def __init__(self, f):
        self.f = f
        self.sub_mouse = WinCtrl.mouse_pub.make_sub()

    def __call__(self, *args, **kwargs):
        self.f(self.sub_mouse, *args, **kwargs)


class mouse_loop_thread(object):  # NOSONAR

    def __init__(self, f, run_when_no_events=False, fps=60):
        self.f = f
        self.sub_mouse = WinCtrl.mouse_pub.make_sub()
        self.sub_cmd = WinCtrl.win_cmd_pub.make_sub()
        self.sub_cmd.return_on_no_data = ''
        self.run_when_no_events = run_when_no_events
        self.fps = fps

    def __call__(self, *args, **kwargs):
        msg_cmd = ''
        while msg_cmd != 'quit':
            mouse_xyzclick = self.sub_mouse.get(blocking=True)  # type: MouseEvent
            if mouse_xyzclick is not self.sub_mouse.return_on_no_data:
                self.f(mouse_xyzclick, *args, **kwargs)
            elif self.run_when_no_events:
                self.f(None, *args, **kwargs)
            msg_cmd = self.sub_cmd.get()
            time.sleep(1.0 / self.fps)
        WinCtrl.quit(force_all_read=False)


class mouse_loop(object):  # NOSONAR

    def __init__(self, f, run_when_no_events=False):
        self.t = threading.Thread(target=mouse_loop_thread(f, run_when_no_events))
        self.t.start()

    def __call__(self, *args, **kwargs):
        return self.t


class key_thread(object):  # NOSONAR

    def __init__(self, f):
        self.f = f
        self.sub_key = WinCtrl.key_pub.make_sub()

    def __call__(self, *args, **kwargs):
        self.f(self.sub_key, *args, **kwargs)


class key_loop_thread(object):  # NOSONAR

    def __init__(self, f, run_when_no_events=False, fps=60):
        self.f = f
        self.sub_key = WinCtrl.key_pub.make_sub()
        self.sub_cmd = WinCtrl.win_cmd_pub.make_sub()
        self.sub_cmd.return_on_no_data = ''
        self.run_when_no_events = run_when_no_events
        self.fps = fps

    def __call__(self, *args, **kwargs):
        msg_cmd = ''
        while msg_cmd != 'quit':
            key_chr = self.sub_key.get()  # type: chr
            if key_chr is not self.sub_key.return_on_no_data:
                self.f(key_chr, *args, **kwargs)
            elif self.run_when_no_events:
                self.f(None, *args, **kwargs)
            msg_cmd = self.sub_cmd.get()
            time.sleep(1.0 / self.fps)
        WinCtrl.quit(force_all_read=False)


class key_loop(object):  # NOSONAR

    def __init__(self,
                 f,  # type: Callable[[chr],None]
                 run_when_no_events=False):
        self.t = threading.Thread(target=key_loop_thread(f, run_when_no_events))
        self.t.start()

    def __call__(self, *args, **kwargs):
        return self.t



import threading
import unittest as ut

import cvpubsubs.webcam_pub as w
from cvpubsubs.window_sub import SubscriberWindows
from cvpubsubs.window_sub.winctrl import WinCtrl
from cvpubsubs import display
from cvpubsubs.input import mouse_loop, key_loop
import numpy as np

if False:
    import numpy as np
    from cvpubsubs.window_sub.mouse_event import MouseEvent


class TestSubWin(ut.TestCase):

    def test_mouse_loop(self):
        @mouse_loop
        def print_mouse_thread(mouse_event):
            print(mouse_event)

        w.VideoHandlerThread().display()

    def test_key_loop(self):
        @key_loop
        def print_key_thread(key_chr):
            print("key pressed: " + str(key_chr))

        w.VideoHandlerThread().display()

    def test_sub(self):
        w.VideoHandlerThread().display()

    def test_image(self):
        img = np.random.uniform(0, 1, (300, 300, 3))
        w.VideoHandlerThread(video_source=img).display()

    def test_image_args(self):
        img = np.random.uniform(0, 1, (30, 30, 3))
        w.VideoHandlerThread(video_source=img, request_size=(300, -1)).display()

    def test_sub_with_args(self):
        video_thread = w.VideoHandlerThread(video_source=0,
                                            request_size=(800, 600),
                                            high_speed=False,
                                            fps_limit=8
                                            )

        video_thread.display()

    def test_sub_with_callback(self):
        def redden_frame_print_spam(frame):
            frame[:, :, 0] = 0
            frame[:, :, 2] = 0

        w.VideoHandlerThread(callbacks=redden_frame_print_spam).display()

    def test_sub_with_callback_exception(self):
        def redden_frame_print_spam(frame):
            frame[:, :, 0] = 0
            frame[:, :, 2] = 1 / 0

        with self.assertRaises(ZeroDivisionError) as e:
            v = w.VideoHandlerThread(callbacks=redden_frame_print_spam)
            v.display()
            self.assertEqual(v.exception_raised, e)

    def test_multi_cams_one_source(self):
        display(0, window_names=['cammy','cammy2'], blocking=True)

    def test_multi_cams_multi_source(self):
        display(0, np.random.uniform(0.0, 1.0, (500,500)), blocking=True)

    def test_nested_frames(self):
        def nest_frame(frame):
            frame = np.asarray([[[[[[frame]]]]], [[[[[frame]]], [[[frame]]]]]])
            return frame

        display(0, callbacks=nest_frame, window_names=["1", "2", "3"], blocking=True)

    def test_nested_frames_exception(self):
        def nest_frame(frame):
            frame = np.asarray([[[[[[frame + 1 / 0]]]]], [[[[[frame]]], [[[frame]]]]]])
            return frame

        v = w.VideoHandlerThread(callbacks=[nest_frame] + w.display_callbacks)
        v.start()

        with self.assertRaises(ZeroDivisionError) as e:
            SubscriberWindows(window_names=[str(i) for i in range(3)],
                              video_sources=[str(0)]
                              ).loop()
            self.assertEqual(v.exception_raised, e)

        v.join()

    def test_conway_life(self):
        from cvpubsubs.webcam_pub import VideoHandlerThread
        from cvpubsubs.callbacks import function_display_callback
        import numpy as np
        import cv2
        img = np.zeros((50, 50, 1))
        img[0:5, 0:5, :] = 1

        def conway(array, coords, finished):
            neighbors = np.sum(array[max(coords[0] - 1, 0):min(coords[0] + 2, 50),
                               max(coords[1] - 1, 0):min(coords[1] + 2, 50)])
            neighbors = max(neighbors - np.sum(array[coords[0:2]]), 0.0)
            if array[coords] == 1.0:
                if neighbors < 2 or neighbors > 3:
                    array[coords] = 0.0
                elif 2 <= neighbors <= 3:
                    array[coords] = 1.0
            else:
                if neighbors == 3:
                    array[coords] = 1.0

        @mouse_loop
        def conway_add(mouse_event  # type:MouseEvent
                       ):
            if 0 <= mouse_event.x < 50 and 0 <= mouse_event.y < 50:
                if mouse_event.flags == cv2.EVENT_FLAG_LBUTTON:
                    img[mouse_event.y - 5:mouse_event.y + 10, mouse_event.x - 5:mouse_event.x + 10, :] = 0.0
                elif mouse_event.flags == cv2.EVENT_FLAG_RBUTTON:
                    img[mouse_event.y - 5:mouse_event.y + 10, mouse_event.x - 5:mouse_event.x + 10, :] = 1.0

        VideoHandlerThread(video_source=img, callbacks=function_display_callback(conway)).display()

    def test_double_win(self):
        vid1 = np.ones((100, 100))
        vid2 = np.zeros((100, 100))
        t1 = w.VideoHandlerThread(vid1)
        t2 = w.VideoHandlerThread(vid2)
        t1.start()
        t2.start()
        SubscriberWindows(window_names=['cammy', 'cammy2'],
                          video_sources=[vid1, vid2]
                          ).loop()
        t1.join()
        t1.join()

    def test_display(self):
        display(np.ones((100, 100)), np.zeros((100, 100)), blocking=True)



import cv2

if False:
    from typing import List


def get_cam_ids():  # type: () -> List[int]
    cam_list = []

    while True:
        cam = cv2.VideoCapture(len(cam_list))
        if not cam.isOpened():
            break
        cam_list.append(len(cam_list))

    return cam_list



import unittest as ut

class TestSubWin(ut.TestCase):

    def test_display_numpy(self):
        from cvpubsubs import display
        import numpy as np

        display(np.random.normal(0.5, .1, (500,500,3)))

    def test_display_numpy_callback(self):
        from cvpubsubs import display
        import numpy as np

        arr = np.random.normal(0.5, .1, (500, 500, 3))

        def fix_arr_cv(arr_in):
            arr_in[:] += np.random.normal(0.01, .005, (500, 500, 3))
            arr_in%=1.0

        display(arr, callbacks= fix_arr_cv, blocking=True)

    def test_display_numpy_loop(self):
        from cvpubsubs import display
        import numpy as np

        arr = np.random.normal(0.5, .1, (500, 500, 3))

        displayer, ids = display(arr, blocking = False)

        while True:
            arr[:] += np.random.normal(0.01, .005, (500, 500, 3))
            arr %= 1.0
            displayer.update(arr, ids[0])
        displayer.end()

    def test_display_tensorflow(self):
        from cvpubsubs import display
        import numpy as np
        from tensorflow.keras import layers, models
        import tensorflow as tf

        for gpu in tf.config.experimental.list_physical_devices("GPU"):
            tf.compat.v2.config.experimental.set_memory_growth(gpu, True)
        #tf.keras.backend.set_floatx("float16")

        displayer, ids = display(0, blocking = False)
        displayer.wait_for_init()
        autoencoder = models.Sequential()
        autoencoder.add(
            layers.Conv2D(20, (3, 3), activation="sigmoid", input_shape=displayer.frames[0].shape)
        )
        autoencoder.add(layers.Conv2DTranspose(3, (3, 3), activation="sigmoid"))

        autoencoder.compile(loss="mse", optimizer="adam")

        while True:
            grab = tf.convert_to_tensor(displayer.frame_dict['0frame'][np.newaxis, ...].astype(np.float32)/255.0)
            autoencoder.fit(grab, grab, steps_per_epoch=1, epochs=1)
            output_image = autoencoder.predict(grab, steps=1)
            displayer.update((output_image[0]*255.0).astype(np.uint8), "uid for autoencoder output")



from cvpubsubs.window_sub.winctrl import WinCtrl
import numpy as np

if False:
    from typing import Union


def global_cv_display_callback(frame,  # type: np.ndarray
                               cam_id  # type: Union[int, str]
                               ):
    from cvpubsubs.window_sub import SubscriberWindows
    """Default callback for sending frames to the global frame dictionary.

    :param frame: The video or image frame
    :type frame: np.ndarray
    :param cam_id: The video or image source
    :type cam_id: Union[int, str]
    """
    SubscriberWindows.frame_dict[str(cam_id) + "frame"] = frame


class function_display_callback(object):  # NOSONAR
    def __init__(self, display_function, finish_function=None):
        """Used for running arbitrary functions on pixels.

        >>> import random
        >>> from cvpubsubs.webcam_pub import VideoHandlerThread
        >>> img = np.zeros((300, 300, 3))
        >>> def fun(array, coords, finished):
        ...     r,g,b = random.random()/20.0, random.random()/20.0, random.random()/20.0
        ...     array[coords[0:2]] = (array[coords[0:2]] + [r,g,b])%1.0
        >>> VideoHandlerThread(video_source=img, callbacks=function_display_callback(fun)).display()

        :param display_function: a function to run on the input image.
        :param finish_function: a function to run on the input image when the other function finishes.
        """
        self.looping = True
        self.first_call = True

        def _run_finisher(self, frame, finished, *args, **kwargs):
            if not callable(finish_function):
                WinCtrl.quit()
            else:
                finished = finish_function(frame, Ellipsis, finished, *args, **kwargs)
                if finished:
                    WinCtrl.quit()

        def _display_internal(self, frame, *args, **kwargs):
            finished = True
            if self.first_call:
                # return to display initial frame
                self.first_call = False
                return
            if self.looping:
                it = np.nditer(frame, flags=['multi_index'])
                while not it.finished:
                    x, y, c = it.multi_index
                    finished = display_function(frame, (x, y, c), finished, *args, **kwargs)
                    it.iternext()
            if finished:
                self.looping = False
                _run_finisher(self, frame, finished, *args, **kwargs)

        self.inner_function = _display_internal

    def __call__(self, *args, **kwargs):
        return self.inner_function(self, *args, **kwargs)



import threading

import numpy as np

from cvpubsubs.webcam_pub.pub_cam import pub_cam_thread
from cvpubsubs.webcam_pub.camctrl import CamCtrl
from cvpubsubs.window_sub.winctrl import WinCtrl
from cvpubsubs.serialize import uid_for_source

if False:
    from typing import Union, Tuple, Any, Callable, List, Optional

    FrameCallable = Callable[[np.ndarray, int], Optional[np.ndarray]]

from cvpubsubs.callbacks import global_cv_display_callback

display_callbacks = [global_cv_display_callback]


class VideoHandlerThread(threading.Thread):
    "Thread for publishing frames from a video source."

    def __init__(self, video_source=0,  # type: Union[int, str, np.ndarray]
                 callbacks=(),  # type: Union[List[FrameCallable], FrameCallable]
                 request_size=(-1, -1),  # type: Tuple[int, int]
                 high_speed=True,  # type: bool
                 fps_limit=240  # type: float
                 ):
        """Sets up the main thread loop.

        :param video_source: The video or image source. Integers typically access webcams, while strings access files.
        :type video_source: Union[int, str]
        :param callbacks: A list of operations to be performed on every frame, including publishing.
        :type callbacks: List[Callable[[np.ndarray, int], Any]]
        :param request_size: Requested video size. Actual size may vary, since this is requesting from the hardware.
        :type request_size: Tuple[int, int]
        :param high_speed: If true, use compression to increase speed.
        :type high_speed: bool
        :param fps_limit: Limits frames per second.
        :type fps_limit: float
        """
        super(VideoHandlerThread, self).__init__(target=self.loop, args=())
        self.cam_id = uid_for_source(video_source)
        self.video_source = video_source
        if callable(callbacks):
            self.callbacks = [callbacks]
        else:
            self.callbacks = callbacks
        self.request_size = request_size
        self.high_speed = high_speed
        self.fps_limit = fps_limit
        self.exception_raised = None

    def loop(self):
        """Continually gets frames from the video publisher, runs callbacks on them, and listens to commands."""
        t = pub_cam_thread(self.video_source, self.request_size, self.high_speed, self.fps_limit)
        while str(self.cam_id) not in CamCtrl.cv_cams_dict:
            continue
        sub_cam = CamCtrl.cam_frame_sub(str(self.cam_id))
        sub_owner = CamCtrl.handler_cmd_sub(str(self.cam_id))
        msg_owner = sub_owner.return_on_no_data = ''
        while msg_owner != 'quit':
            frame = sub_cam.get(blocking=True, timeout=1.0)  # type: np.ndarray
            if frame is not None:
                frame_c = None
                for c in self.callbacks:
                    try:
                        frame_c = c(frame)
                    except TypeError as te:
                        raise TypeError("Callback functions for cvpubsub need to accept two arguments: array and uid")
                    except Exception as e:
                        self.exception_raised = e
                        frame = frame_c = self.exception_raised
                        CamCtrl.stop_cam(self.cam_id)
                        WinCtrl.quit()
                        raise e
                if frame_c is not None:
                    global_cv_display_callback(frame_c, self.cam_id)
                else:
                    global_cv_display_callback(frame, self.cam_id)
            msg_owner = sub_owner.get()
        sub_owner.release()
        sub_cam.release()
        CamCtrl.stop_cam(self.cam_id)
        t.join()

    def display(self,
                callbacks=()  # type: List[Callable[[List[np.ndarray]], Any]]
                ):
        from cvpubsubs.window_sub import SubscriberWindows

        """Default display operation. For multiple video sources, please use something outside of this class.

        :param callbacks: List of callbacks to be run on frames before displaying to the screen.
        :type callbacks: List[Callable[[List[np.ndarray]], Any]]
        """
        self.start()
        SubscriberWindows(video_sources=[self.cam_id], callbacks=callbacks).loop()
        self.join()
        if self.exception_raised is not None:
            raise self.exception_raised



import cvpubsubs.webcam_pub as w
import unittest as ut


class TestFrameHandler(ut.TestCase):
    i = 0

    def test_handler(self):

        def test_frame_handler(frame, cam_id):
            if self.i == 200:
                w.CamCtrl.stop_cam(cam_id)
            if self.i % 100 == 0:
                print(frame.shape)
            self.i += 1

        w.VideoHandlerThread(0, [test_frame_handler],
                               request_size=(1280, 720),
                               high_speed=True,
                               fps_limit=240)
